"""Utility functions for feature hashing that encodes a feature value to a vector.
"""

import mmh3
import numpy as np


def n_feature_hash(feature, dims, seeds):
    """N-hot-encoded feature hashing.

    Parameters
    ----------
    feature : str
        Target feature value represented as string.

    dims : list of int
        Number of dimensions for each hash value.

    seeds : list of float)
        Seed of each hash function (MurmurHash3).

    Returns
    -------
    array
        n-hot-encoded vector for ``feature``.
    """
    vec = np.zeros(sum(dims))
    offset = 0

    for seed, dim in zip(seeds, dims):
        vec[offset : (offset + dim)] = feature_hash(feature, dim, seed)
        offset += dim

    return vec


def feature_hash(feature, dim, seed=123):
    """Onehot-encoded Feature hashing.

    Parameters
    ----------
    feature : str
        Target feature value represented as string.

    dim : int
        Number of dimensions for a hash value.

    seed : float
        Seed of a MurmurHash3 hash function.

    Returns
    -------
    array
        Onehot-encoded vector for ``feature``.
    """
    vec = np.zeros(dim)
    i = mmh3.hash(feature, seed) % dim
    vec[i] = 1
    return vec


def multiple_feature_hash(feature, dim, seed=123):
    """Onehot-encoded feature hashing using multiple hash functions.
    This technique is effective to prevent collisions.

    Parameters
    ----------
    feature : str
        Target feature value represented as string.

    dim : int
        Number of dimensions for a hash value.

    seed : float
        Seed of a MurmurHash3 hash function.

    Returns
    -------
    array
        Onehot-encoded vector for ``feature``.
    """
    vec = np.zeros(dim)
    i = mmh3.hash(feature, seed) % dim
    vec[i] = 1 if mmh3.hash(feature) % 2 else -1
    return vec



from ..base import RecommenderMixin
from ..model import MatrixFactorization
from .. import logger

import numpy as np


class MFRecommender(MatrixFactorization, RecommenderMixin):

    __doc__ = MatrixFactorization.__doc__

    def initialize(self, static=False):
        """Initialize a recommender by resetting stored users and items.
        Default the number of users and items to zero.

        Parameters
        ----------
        static : bool, default=False
            Disable incremental update if ``True``s.
        """
        super(MFRecommender, self).initialize()

        # if True, parameters will not be updated in evaluation
        self.static = static

    def register_user(self, user):
        super(MFRecommender, self).register_user(user)
        self.users[user.index]["vec"] = np.random.normal(0.0, 0.1, self.k)

    def register_item(self, item):
        super(MFRecommender, self).register_item(item)
        i_vec = np.random.normal(0.0, 0.1, (1, self.k))
        if self.Q.size == 0:
            self.Q = i_vec
        else:
            self.Q = np.concatenate((self.Q, i_vec))

    def update(self, e, batch_train=False):
        # static baseline; w/o updating the model
        if not batch_train and self.static:
            return

        if e.value != 1.0:
            logger.info(
                "Incremental matrix factorization assumes implicit feedback"
                " recommendation, so the event value is automatically converted"
                " into 1.0"
            )
            e.value = 1.0

        self.update_model(e.user.index, e.item.index, e.value)

    def score(self, user, candidates):
        pred = np.dot(self.users[user.index]["vec"], self.Q[candidates, :].T)
        return np.abs(1.0 - pred.flatten())

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates)



"""MovieLens datasets created by GroupLens."""

from ..data.entity import User, Item, Event

import os
import time
import numpy as np
from calendar import monthrange
from datetime import datetime, timedelta

from sklearn.utils import Bunch


def load_movies(data_home, size):
    """Load movie genres as a context.

    Parameters
    ----------
    data_home : str
        Absolute path to MovieLens data folder.

    size : {"100k", "1m"}
        String notation of MovieLens data size.

    Returns
    -------
    dict of movie vectors
        item_id -> numpy array (n_genre, ).
    """
    all_genres = [
        "Action",
        "Adventure",
        "Animation",
        "Children's",
        "Comedy",
        "Crime",
        "Documentary",
        "Drama",
        "Fantasy",
        "Film-Noir",
        "Horror",
        "Musical",
        "Mystery",
        "Romance",
        "Sci-Fi",
        "Thriller",
        "War",
        "Western",
    ]
    n_genre = len(all_genres)

    movies = {}

    if size == "100k":
        with open(os.path.join(data_home, "u.item"), encoding="ISO-8859-1") as f:
            lines = list(map(lambda l: l.rstrip().split("|"), f.readlines()))

        for line in lines:
            movie_vec = np.zeros(n_genre)
            for i, flg_chr in enumerate(line[-n_genre:]):
                if flg_chr == "1":
                    movie_vec[i] = 1.0
            movie_id = int(line[0])
            movies[movie_id] = movie_vec
    elif size == "1m":
        with open(os.path.join(data_home, "movies.dat"), encoding="ISO-8859-1") as f:
            lines = list(map(lambda l: l.rstrip().split("::"), f.readlines()))

        for item_id_str, title, genres in lines:
            movie_vec = np.zeros(n_genre)
            for genre in genres.split("|"):
                i = all_genres.index(genre)
                movie_vec[i] = 1.0
            item_id = int(item_id_str)
            movies[item_id] = movie_vec

    return movies


def load_users(data_home, size):
    """Load user demographics as contexts.
    User ID -> {sex (M/F), age (7 groupd), occupation(0-20; 21)}


    Parameters
    ----------
    data_home : str
        Absolute path to MovieLens data folder.

    size : {"100k", "1m"}
        String notation of MovieLens data size.

    Returns
    -------
    dict of user vectors
        user_id -> numpy array (1 + 1 + 21, );
        (sex_flg + age_group + n_occupation, ).
    """
    ages = [1, 18, 25, 35, 45, 50, 56, 999]

    users = {}

    if size == "100k":
        all_occupations = [
            "administrator",
            "artist",
            "doctor",
            "educator",
            "engineer",
            "entertainment",
            "executive",
            "healthcare",
            "homemaker",
            "lawyer",
            "librarian",
            "marketing",
            "none",
            "other",
            "programmer",
            "retired",
            "salesman",
            "scientist",
            "student",
            "technician",
            "writer",
        ]

        with open(os.path.join(data_home, "u.user"), encoding="ISO-8859-1") as f:
            lines = list(map(lambda l: l.rstrip().split("|"), f.readlines()))

        for user_id_str, age_str, sex_str, occupation_str, zip_code in lines:
            user_vec = np.zeros(1 + 1 + 21)  # 1 categorical, 1 value, 21 categorical
            user_vec[0] = 0 if sex_str == "M" else 1  # sex

            # age (ML1M is "age group", but 100k has actual "age")
            age = int(age_str)
            for i in range(7):
                if age >= ages[i] and age < ages[i + 1]:
                    user_vec[1] = i
                    break

            user_vec[
                2 + all_occupations.index(occupation_str)
            ] = 1  # occupation (1-of-21)
            users[int(user_id_str)] = user_vec
    elif size == "1m":
        with open(os.path.join(data_home, "users.dat"), encoding="ISO-8859-1") as f:
            lines = list(map(lambda l: l.rstrip().split("::"), f.readlines()))

        for user_id_str, sex_str, age_str, occupation_str, zip_code in lines:
            user_vec = np.zeros(1 + 1 + 21)  # 1 categorical, 1 value, 21 categorical
            user_vec[0] = 0 if sex_str == "M" else 1  # sex
            user_vec[1] = ages.index(int(age_str))  # age group (1, 18, ...)
            user_vec[2 + int(occupation_str)] = 1  # occupation (1-of-21)
            users[int(user_id_str)] = user_vec

    return users


def load_ratings(data_home, size):
    """Load all samples in the dataset.

    Parameters
    ----------
    data_home : str
        Absolute path to MovieLens data folder.

    size : {"100k", "1m"}
        String notation of MovieLens data size.

    Returns
    -------
    array
        Single row is ``[user_id, item_id, rating, timestamp]``.
        Rows are sorted by ``timestamp.
    """

    if size == "100k":
        with open(os.path.join(data_home, "u.data"), encoding="ISO-8859-1") as f:
            lines = list(
                map(lambda l: list(map(int, l.rstrip().split("\t"))), f.readlines())
            )
    elif size == "1m":
        with open(os.path.join(data_home, "ratings.dat"), encoding="ISO-8859-1") as f:
            lines = list(
                map(lambda l: list(map(int, l.rstrip().split("::"))), f.readlines())
            )

    ratings = []

    for l in lines:
        # Since we consider positive-only feedback setting,
        # ratings < 5 will be excluded.
        if l[2] == 5:
            ratings.append(l)

    ratings = np.asarray(ratings)

    # sorted by timestamp
    return ratings[np.argsort(ratings[:, 3])]


def delta(d1, d2, opt="d"):
    """Compute difference between given 2 dates in month/day.

    Parameters
    ----------
    d1 : datetime
        First date.

    d2 : datetime
        Second date.

    opt : {"d", "m"}, default="d"
        ``"d"`` if `d1` and `d2` have day-level granularity. If they only tell about
        month (day is always 1), ``"m"`` interpolates the information.

    Returns
    -------
    int
        Difference between two dates in days.
    """
    delta = 0

    if opt == "m":
        while True:
            mdays = monthrange(d1.year, d1.month)[1]
            d1 += timedelta(days=mdays)
            if d1 <= d2:
                delta += 1
            else:
                break
    else:
        delta = (d2 - d1).days

    return delta


def fetch_movielens(data_home=None, size="100k"):
    """Fetch MovieLens data in the form of `User`, `Item`, `Event`.

    Parameters
    ----------
    data_home : str
        Absolute path to MovieLens data folder.

    size : {"100k", "1m"}, default="100k"
        String notation of MovieLens data size.

    Returns
    -------
    sklearn.utils.Bunch
        Dictionary-like object, with the following attributes.

        samples : list of Event
            All rating events.
        can_repeat : bool
            ``False`` because MovieLens does not contain same user-item interaction
            more than twice.
        contexts : dict
            Contextual feature name -> Number of its dimensions.
        n_user : int
            Number of unique users.
        n_item : int
            Number of unique items.
        n_sample : int
            Number of events.
    """
    assert data_home is not None

    if size not in ("100k", "1m"):
        raise ValueError("size can only be '100k' or '1m', got %s" % size)

    ratings = load_ratings(data_home, size)
    users = load_users(data_home, size)
    movies = load_movies(data_home, size)

    samples = []

    user_ids = {}
    item_ids = {}

    head_date = datetime(*time.localtime(ratings[0, 3])[:6])
    dts = []

    last = {}

    for user_id, item_id, rating, timestamp in ratings:
        # give an unique user index
        if user_id in user_ids:
            u_index = user_ids[user_id]
        else:
            u_index = len(user_ids)
            user_ids[user_id] = u_index

        # give an unique item index
        if item_id in item_ids:
            i_index = item_ids[item_id]
        else:
            i_index = len(item_ids)
            item_ids[item_id] = i_index

        # delta days
        date = datetime(*time.localtime(timestamp)[:6])
        dt = delta(head_date, date)
        dts.append(dt)

        weekday_vec = np.zeros(7)
        weekday_vec[date.weekday()] = 1

        if user_id in last:
            last_item_vec = last[user_id]["item"]
            last_weekday_vec = last[user_id]["weekday"]
        else:
            last_item_vec = np.zeros(18)
            last_weekday_vec = np.zeros(7)

        others = np.concatenate((weekday_vec, last_item_vec, last_weekday_vec))

        user = User(u_index, users[user_id])
        item = Item(i_index, movies[item_id])

        sample = Event(user, item, 1.0, others)
        samples.append(sample)

        # record users' last rated movie features
        last[user_id] = {"item": movies[item_id], "weekday": weekday_vec}

    # contexts in this dataset
    # 1 delta time, 18 genres, and 23 demographics
    # (1 for M/F, 1 for age, 21 for occupation(0-20))
    # 7 for day of week, 18 for the last rated item genres, 7 for the last day of week
    return Bunch(
        samples=samples,
        can_repeat=False,
        contexts={"others": 7 + 18 + 7, "item": 18, "user": 23},
        n_user=len(user_ids),
        n_item=len(item_ids),
        n_sample=len(samples),
    )



from sklearn.base import BaseEstimator

from ..utils.projection import (
    Raw,
    RandomProjection,
    RandomMaclaurinProjection,
    TensorSketchProjection,
)

import numpy as np
import numpy.linalg as ln
import scipy.sparse as sp
from sklearn import preprocessing


class OnlineSketch(BaseEstimator):

    r"""Online matrix sketching.

    Parameters
    ----------
    p : int
        Number of dimensions of an input vector.

    k : int, default=40
        Number of dimensions of a projected vector.

    ell : int, default=-1
        Upper bound for the number of tracked orthogonal bases :math:`\ell`.
        Defaults to :math:`\sqrt{k}` if the value is negative.

    r : int, default=-1
        Number of tracked orthogonal bases. Defaults to :math:`\ell / 2` (ceiled).

    proj : str, default='Raw'
        Vector projection method for dimensionality reduction. Options are {'Raw',
        'RandomProjection', 'RandomMaclaurinProjection', 'TensorSketchProjection'}.
        See `flurs.utils.projection` for the detail.

    References
    ----------
    .. [1] T. Kitazawa.
           `Sketching Dynamic User-Item Interactions for Online Item Recommendation
           <http://dl.acm.org/citation.cfm?id=3022152>`_.
           In *Proc. of CHIIR 2017*, March 2017.
    """

    def __init__(self, p=None, k=40, ell=-1, r=-1, proj="Raw"):
        assert p is not None

        # number of dimensions of input vectors
        self.p = p

        # dimension of projected vectors
        # for `Raw` (i.e. w/o projection), k must equat to p
        self.k = self.p if proj == "Raw" else k

        # if there is no preference for ell,
        # this will be sqrt(k) similarly to what the original streaming anomaly
        # detection paper did
        self.ell = int(np.sqrt(self.k)) if ell < 1 else ell

        # number of tracked orthogonal bases
        # * upper bound of r is ell (r <= ell) because U_r is obtained from
        #   SVD(B) (or SVD(E)), and B and E always have ell columns
        self.r = int(np.ceil(self.ell / 2)) if r < 1 else np.min(r, self.ell)

        # initialize projection instance which is specified by `proj` argument
        if proj == "Raw":
            self.proj = Raw(self.k, self.p)
        elif proj == "RandomProjection":
            self.proj = RandomProjection(self.k, self.p)
        elif proj == "RandomMaclaurinProjection":
            self.proj = RandomMaclaurinProjection(self.k, self.p)
        elif proj == "TensorSketchProjection":
            self.proj = TensorSketchProjection(self.k, self.p)

        self.i_mat = sp.csr_matrix([])

    def update_model(self, y):
        y = self.proj.reduce(np.array([y]).T)
        y = np.ravel(preprocessing.normalize(y, norm="l2", axis=0))

        if not hasattr(self, "B"):
            self.B = np.zeros((self.k, self.ell))

        # combine current sketched matrix with input at time t
        zero_cols = np.nonzero(
            [np.isclose(s_col, 0.0) for s_col in np.sum(self.B, axis=0)]
        )[0]
        j = (
            zero_cols[0] if zero_cols.size != 0 else self.ell - 1
        )  # left-most all-zero column in B
        self.B[:, j] = y

        U, s, V = ln.svd(self.B, full_matrices=False)

        # update the tracked orthonormal bases
        self.U_r = U[:, : self.r]

        # update ell orthogonal bases
        U_ell = U[:, : self.ell]
        s_ell = s[: self.ell]

        # shrink step in the Frequent Directions algorithm
        # (shrink singular values based on the squared smallest singular value)
        delta = s_ell[-1] ** 2
        s_ell = np.sqrt(s_ell ** 2 - delta)

        self.B = np.dot(U_ell, np.diag(s_ell))


class OnlineRandomSketch(OnlineSketch):

    """Inspired by: Streaming Anomaly Detection using Randomized Matrix Sketching
    [WIP] many matrix multiplications are computational heavy
    """

    def update_model(self, y):
        y = self.proj.reduce(np.array([y]).T)
        y = np.ravel(preprocessing.normalize(y, norm="l2", axis=0))

        if not hasattr(self, "E"):
            self.E = np.zeros((self.k, self.ell))

        # combine current sketched matrix with input at time t
        zero_cols = np.nonzero(
            [np.isclose(s_col, 0.0) for s_col in np.sum(self.E, axis=0)]
        )[0]
        j = (
            zero_cols[0] if zero_cols.size != 0 else self.ell - 1
        )  # left-most all-zero column in B
        self.E[:, j] = y

        Gaussian = np.random.normal(0.0, 0.1, (self.k, 100 * self.ell))
        MM = np.dot(self.E, self.E.T)
        Q, R = ln.qr(np.dot(MM, Gaussian))

        # eig() returns eigen values/vectors with unsorted order
        s, A = ln.eig(np.dot(np.dot(Q.T, MM), Q))
        order = np.argsort(s)[::-1]
        s = s[order]
        A = A[:, order]

        U = np.dot(Q, A)

        # update the tracked orthonormal bases
        self.U_r = U[:, : self.r]

        # update ell orthogonal bases
        U_ell = U[:, : self.ell]
        s_ell = s[: self.ell]

        # shrink step in the Frequent Directions algorithm
        delta = s_ell[-1]
        s_ell = np.sqrt(s_ell - delta)

        self.E = np.dot(U_ell, np.diag(s_ell))


class OnlineSparseSketch(OnlineSketch):

    """Inspired by: Efficient Frequent Directions Algorithm for Sparse Matrices"""

    def update_model(self, y):
        y = self.proj.reduce(np.array([y]).T)
        y = preprocessing.normalize(y, norm="l2", axis=0)  # (k, 1)

        if not hasattr(self, "B"):
            self.p_failure = 0.1
            self.B = np.zeros((self.k, self.ell))
            self.A = np.array([])

        U, s, V = ln.svd(self.B, full_matrices=False)

        # update the tracked orthonormal bases
        self.U_r = U[:, : self.r]

        if self.A.size == 0:
            self.A = np.empty_like(y)
            self.A[:] = y[:]
        else:
            self.A = np.concatenate((self.A, y), axis=1)

        if np.count_nonzero(self.A) >= (self.ell * self.k) or self.A.shape[1] == self.k:
            B = self.__boosted_sparse_shrink(self.A, self.ell, self.p_failure)
            self.B = self.__dense_shrink(np.concatenate((self.B, B), axis=1), self.ell)
            self.A = np.array([])

    def __simultaneous_iteration(self, A, k, eps):
        n, d = A.shape

        q = int(np.log((n / eps) / eps))
        G = np.random.normal(0.0, 1.0, (d, k))

        # Gram-Schmidt
        Y = np.dot(np.dot(ln.matrix_power(np.dot(A, A.T), q), A), G)  # (n, k)
        Q, R = ln.qr(Y, mode="complete")

        return Q

    def __sparse_shrink(self, A, ell):
        m, d = A.shape

        Z = self.__simultaneous_iteration(A, ell, 1 / 4)
        P = sp.csr_matrix(np.dot(Z.T, A))

        # 1 <= ell < min(P.shape)
        H_ell, s_ell, V_ell = sp.linalg.svds(P, ell)

        # shrink step in the Frequent Directions algorithm
        # (shrink singular values based on the squared smallest singular value)
        lambda_ell = s_ell[-1] ** 2
        s_ell = np.sqrt(s_ell ** 2 - lambda_ell)

        return np.dot(H_ell, np.diag(s_ell))

    def __boosted_sparse_shrink(self, A, ell, p_failure):
        al = (6 / 41) * ell
        sq_norm_A = ln.norm(A, ord="fro") ** 2
        AA = np.dot(A, A.T)

        while True:
            B = self.__sparse_shrink(A, ell)
            delta = (sq_norm_A - (ln.norm(B, ord="fro") ** 2)) / al

            C = (AA - np.dot(B, B.T)) / (delta / 2)
            if self.__verify_spectral(C, p_failure):
                return B

    def __dense_shrink(self, A, ell):
        H, s, V = ln.svd(A, full_matrices=False)

        H_ell = H[:, :ell]
        s_ell = s[:ell]

        # shrink step in the Frequent Directions algorithm
        # (shrink singular values based on the squared smallest singular value)
        lambda_ell = s_ell[-1] ** 2
        s_ell = np.sqrt(s_ell ** 2 - lambda_ell)

        return np.dot(H_ell, np.diag(s_ell))

    def __verify_spectral(self, C, p_failure):
        d = C.shape[0]
        c = 1  # some constant

        if not hasattr(self, "i_verify"):
            self.i_verify = 0

        self.i_verify += 1
        p_failure_i = p_failure / (2 * (self.i_verify ** 2))

        # pick a point uniformly at ranom from the d-dimensional unit sphere
        # http://stackoverflow.com/questions/15880367/python-uniform-distribution-of-points-on-4-dimensional-sphere
        normal_deviates = np.random.normal(size=(d, 1))
        radius = np.sqrt((normal_deviates ** 2).sum(axis=0))
        x = normal_deviates / radius

        return (
            ln.norm(np.dot(ln.matrix_power(C, int(c * np.log(d / p_failure_i))), x))
            <= 1
        )



from sklearn.base import BaseEstimator

import numpy as np


class MatrixFactorization(BaseEstimator):

    r"""Incremental Matrix Factorization (MF).

    Parameters
    ----------
    k : int, default=40
        Number of latent factors.

    l2_reg : float, default=0.01
        :math:`\lambda` for L2 regularization.

    learn_rate : float, default=0.003
        Learning rate :math:`\eta`.

    References
    ----------
    .. [1] J. Vinagre et al.
           `Fast Incremental Matrix Factorization for Recommendation with Positive-only
           Feedback <http://link.springer.com/chapter/10.1007/978-3-319-08786-3_41>`_.
           In *Proc. of UMAP 2014*, pp. 459-470, July 2014.
    """

    def __init__(self, k=40, l2_reg=0.01, learn_rate=0.003):
        self.k = k
        self.l2_reg_u = l2_reg
        self.l2_reg_i = l2_reg
        self.learn_rate = learn_rate

        self.Q = np.array([])

    def update_model(self, ua, ia, value):
        u_vec = self.users[ua]["vec"]
        i_vec = self.Q[ia]

        err = value - np.inner(u_vec, i_vec)

        grad = -2.0 * (err * i_vec - self.l2_reg_u * u_vec)
        next_u_vec = u_vec - self.learn_rate * grad

        grad = -2.0 * (err * u_vec - self.l2_reg_i * i_vec)
        next_i_vec = i_vec - self.learn_rate * grad

        self.users[ua]["vec"] = next_u_vec
        self.Q[ia] = next_i_vec



"""Ranking-based evaluation metrics for recommender systems"""

import numpy as np


def count_true_positive(truth, recommend):
    """Count number of true positives from given sets of samples.

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    Returns
    -------
    int
        Number of true positives.
    """
    tp = 0
    for r in recommend:
        if r in truth:
            tp += 1
    return tp


def recall(truth, recommend, k=None):
    """Recall@k.

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    k : int or None, default=None
        Top-k items in ``recommend`` are considered to be recommended.
        Defaults to ``len(recommend)``.

    Returns
    -------
    float
        Recall@k.
    """
    if len(truth) == 0:
        if len(recommend) == 0:
            return 1.0
        return 0.0

    if k is None:
        k = len(recommend)
    return count_true_positive(truth, recommend[:k]) / float(truth.size)


def precision(truth, recommend, k=None):
    """Precision@k.

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    k : int or None, default=None
        Top-k items in ``recommend`` are considered to be recommended.
        Defaults to ``len(recommend)``.

    Returns
    -------
    float
        Precision@k.
    """
    if len(recommend) == 0:
        if len(truth) == 0:
            return 1.0
        return 0.0

    if k is None:
        k = len(recommend)
    return count_true_positive(truth, recommend[:k]) / float(k)


def average_precision(truth, recommend):
    """Average Precision (AP).

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    Returns
    -------
    float
        Average Precision.
    """
    if len(truth) == 0:
        if len(recommend) == 0:
            return 1.0
        return 0.0

    tp = accum = 0.0
    for n in range(recommend.size):
        if recommend[n] in truth:
            tp += 1.0
            accum += tp / (n + 1.0)
    return accum / truth.size


def auc(truth, recommend):
    """Area under the ROC curve (AUC).

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    Returns
    -------
    float
        AUC.
    """
    tp = correct = 0.0
    for r in recommend:
        if r in truth:
            # keep track number of true positives placed before
            tp += 1.0
        else:
            correct += tp
    # number of all possible tp-fp pairs
    pairs = tp * (recommend.size - tp)

    # if there is no TP (or no FP), it's meaningless for this metric (i.e., AUC=0.5)
    if pairs == 0:
        return 0.5

    return correct / pairs


def reciprocal_rank(truth, recommend):
    """Reciprocal Rank (RR).

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    Returns
    -------
    float
        Reciprocal Rank.
    """
    for n in range(recommend.size):
        if recommend[n] in truth:
            return 1.0 / (n + 1)
    return 0.0


def mpr(truth, recommend):
    """Mean Percentile Rank (MPR).

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    Returns
    -------
    float
        Mean Percentile Rank.
    """
    if len(recommend) == 0 and len(truth) == 0:
        return 0.0  # best
    elif len(truth) == 0 or len(truth) == 0:
        return 100.0  # worst

    accum = 0.0
    n_recommend = recommend.size
    for t in truth:
        r = np.where(recommend == t)[0][0] / float(n_recommend)
        accum += r
    return accum * 100.0 / truth.size


def ndcg(truth, recommend, k=None):
    """Normalized Discounted Cumulative Grain (NDCG).

    Parameters
    ----------
    truth : numpy 1d array
        Set of truth samples.

    recommend : numpy 1d array
        Ordered listed of recommended samples.

    k : int or None, default=None
        Top-k items in ``recommend`` are considered to be recommended.
        Defaults to ``len(recommend)``.

    Returns
    -------
    float
        NDCG@k.
    """
    if k is None:
        k = len(recommend)

    def idcg(n_possible_truth):
        res = 0.0
        for n in range(n_possible_truth):
            res += 1.0 / np.log2(n + 2)
        return res

    dcg = 0.0
    for n, r in enumerate(recommend[:k]):
        if r not in truth:
            continue
        dcg += 1.0 / np.log2(n + 2)

    res_idcg = idcg(np.min([truth.size, k]))
    if res_idcg == 0.0:
        return 0.0
    return dcg / res_idcg



from ..base import RecommenderMixin
from ..model import UserKNN

import numpy as np


class UserKNNRecommender(UserKNN, RecommenderMixin):

    __doc__ = UserKNN.__doc__

    def initialize(self):
        super(UserKNNRecommender, self).initialize()

    def register_user(self, user):
        super(UserKNNRecommender, self).register_user(user)

        self.R = self.__insert_row(self.R, self.n_item)

        self.S = self.__insert_row(self.S, self.n_user - 1)
        self.S = self.__insert_col(self.S, self.n_user)

        self.B = self.__insert_row(self.B, self.n_user - 1)
        self.B = self.__insert_col(self.B, self.n_user)

        self.C = self.__insert_row(self.C, self.n_user - 1)
        self.C = self.__insert_col(self.C, self.n_user)

        self.D = self.__insert_row(self.D, self.n_user - 1)
        self.D = self.__insert_col(self.D, self.n_user)

        # keep track how many times each user interacted with items
        # to compute user's mean
        self.users[user.index]["count"] = 0

        # current average rating of the user
        self.users[user.index]["mean"] = 0.0

    def register_item(self, item):
        super(UserKNNRecommender, self).register_item(item)
        self.R = self.__insert_col(self.R, self.n_user)

    def update(self, e, batch_train=False):
        self.update_model(e.user.index, e.item.index, e.value)

    def score(self, user, candidates):
        ua = user.index

        # find k nearest neightbors
        top_uys = np.argsort(self.S[ua, :])[::-1][: self.k]

        pred = np.ones(len(candidates)) * self.users[ua]["mean"]
        for pi, ii in enumerate(candidates):
            denom = numer = 0.0
            for uy in top_uys:
                numer += (self.R[uy, ii] - self.users[uy]["mean"]) * self.S[ua, uy]
                denom += self.S[ua, uy]
            if denom != 0.0:
                pred[pi] += numer / denom

        return np.abs(pred)

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates, rev=True)

    def __insert_row(self, mat, n_col):
        row = np.zeros((1, n_col))
        if mat.size == 0:
            mat = row
        else:
            mat = np.concatenate((mat, row))
        return mat

    def __insert_col(self, mat, n_row):
        col = np.zeros((n_row, 1))
        if mat.size == 0:
            mat = col
        else:
            mat = np.concatenate((mat, col), axis=1)
        return mat



from sklearn.base import BaseEstimator
from ..base import RecommenderMixin

import numpy as np


class Random(BaseEstimator, RecommenderMixin):

    """Random baseline that randomly ranks candidates."""

    def __init__(self):
        pass

    def initialize(self):
        super(Random, self).initialize()

    def register_user(self, user):
        super(Random, self).register_user(user)

    def register_item(self, item):
        super(Random, self).register_item(item)

    def update(self, e, batch_train=False):
        pass

    def score(self, user, candidates):
        return np.random.rand(len(candidates))

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates)



from sklearn.base import BaseEstimator

import numpy as np


class BPRMF(BaseEstimator):

    r"""Incremental Matrix Factorization with BPR optimization.

    Parameters
    ----------
    k : int, default=40
        Number of latent factors.

    l2_reg : float, default=0.01
        :math:`\lambda` for L2 regularization.

    learn_rate : float, default=0.003
        Learning rate :math:`\eta`.

    References
    ----------
    .. [1] S. Rendle et al.
           **BPR: Bayesian Personalized Ranking from Implicit Feedback**.
           In *Proc. of UAI 2009*, pp. 452-461, June 2009.
    """

    def __init__(self, k=40, l2_reg=0.01, learn_rate=0.003):
        self.k = k
        self.l2_reg_u = l2_reg
        self.l2_reg_i = l2_reg  # positive items: i
        self.l2_reg_j = l2_reg  # negative items: j
        self.learn_rate = learn_rate

        self.Q = np.array([])

    def update_model(self, ua, ia):
        u_vec = self.users[ua]["vec"]
        i_vec = self.Q[ia]
        x_ui = np.inner(u_vec, i_vec)

        unobserved = list(set(range(self.n_item)) - self.users[ua]["known_items"])

        # choose one negative (i.e., unobserved) sample
        j = np.random.choice(unobserved)

        j_vec = self.Q[j]

        x_uj = np.inner(u_vec, j_vec)
        x_uij = x_ui - x_uj
        sigmoid = np.e ** (-x_uij) / (1 + np.e ** (-x_uij))

        grad = i_vec - j_vec
        next_u_vec = u_vec + self.learn_rate * (sigmoid * grad + self.l2_reg_u * u_vec)

        grad = u_vec
        next_i_vec = i_vec + self.learn_rate * (sigmoid * grad + self.l2_reg_i * i_vec)

        grad = -u_vec
        next_j_vec = j_vec + self.learn_rate * (sigmoid * grad + self.l2_reg_j * j_vec)

        self.users[ua]["vec"] = next_u_vec
        self.Q[ia] = next_i_vec
        self.Q[j] = next_j_vec



from logging import getLogger, StreamHandler, Formatter, DEBUG

logger = getLogger(__name__)
handler = StreamHandler()
handler.setFormatter(
    Formatter("%(asctime)s : %(levelname)s : %(process)d : %(message)s")
)
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)

__version__ = "0.0.5"

__all__ = ["baseline", "data", "datasets", "recommender", "utils"]



"""Utility classes for vector projection.
"""

import numpy as np
import scipy.sparse as sp
from sklearn.utils.extmath import safe_sparse_dot


class BaseProjection(object):

    """Base class for projection of feature vectors.

    Parameters
    ----------
    k : int
        Number of reduced dimensions (i.e. rows of projection mat.).

    p : int
        Number of input dimensions (i.e. cols of projection mat.).
        ``p`` will be increased in the future due to new user/item/context insersion.
    """

    def __init__(self, k, p):
        pass

    def insert_proj_col(self, offset):
        """Insert a new column into a projection matrix.

        Parameters
        ----------
        offset : int
            Column index of where the new dimension is inserted
        """
        pass

    def reduce(self, Y):
        """Apply projection to an input matrix.

        Parameters
        ----------
        Y : numpy array, (p, n)
            Input p-by-n matrix projected to a k-by-n matrix.

        Returns
        -------
        array
            A k-by-n projected matrix.
        """
        return


class Raw(BaseProjection):

    """Raw projector that does nothing regardless of the parameters
    i.e., a projector is an identity matrix.

    Parameters
    ----------
    k : int
        Number of reduced dimensions (i.e. rows of projection mat.).

    p : int
        Number of input dimensions (i.e. cols of projection mat.).
        ``p`` will be increased in the future due to new user/item/context insersion.
    """

    def __init__(self, k, p):
        # k == p
        self.E = np.identity(p)

    def insert_proj_col(self, offset):
        pass

    def reduce(self, Y):
        return safe_sparse_dot(self.E, Y)


class RandomProjection(BaseProjection):

    """Projection based on a randomly initialized matrix.

    Parameters
    ----------
    k : int
        Number of reduced dimensions (i.e. rows of projection mat.).

    p : int
        Number of input dimensions (i.e. cols of projection mat.).
        ``p`` will be increased in the future due to new user/item/context insersion.

    density : float, default=0.2
        Density parameter used to create a projection matrix.

    References
    ----------
    .. [1] D. Achlioptas. Database-friendly random projections:
           Johnson-Lindenstrauss with binary coins.
    .. [2] P. Li, et al. Very sparse random projections.
    .. [3] `Scikit-learn documentation: Sparse random projection
           <http://scikit-learn.org/stable/modules/random_projection.html#sparse-random-projection>`_.
    """

    def __init__(self, k, p, density=0.2):
        self.k = k
        self.density = density
        self.R = sp.csr_matrix(self.__create_proj_mat((k, p)))

    def insert_proj_col(self, offset):
        col = self.__create_proj_mat((self.k, 1))
        R = self.R.toarray()
        self.R = sp.csr_matrix(
            np.concatenate((R[:, :offset], col, R[:, offset:]), axis=1)
        )

    def reduce(self, Y):
        return safe_sparse_dot(self.R, Y)

    def __create_proj_mat(self, size):
        # [1]_
        # return np.random.choice([-np.sqrt(3), 0, np.sqrt(3)],
        #                         size=size, p=[1 / 6, 2 / 3, 1 / 6])

        # [2]_
        s = 1 / self.density
        return np.random.choice(
            [-np.sqrt(s / self.k), 0, np.sqrt(s / self.k)],
            size=size,
            p=[1 / (2 * s), 1 - 1 / s, 1 / (2 * s)],
        )


class RandomMaclaurinProjection(BaseProjection):

    """Random Maclaurin Projection.

    Parameters
    ----------
    k : int
        Number of reduced dimensions (i.e. rows of projection mat.).

    p : int
        Number of input dimensions (i.e. cols of projection mat.).
        ``p`` will be increased in the future due to new user/item/context insersion.
    """

    def __init__(self, k, p):
        self.k = k

        self.W1 = np.random.choice([1, -1], size=(k, p))
        self.W2 = np.random.choice([1, -1], size=(k, p))

    def insert_proj_col(self, offset):
        col = np.random.choice([1, -1], size=(self.k, 1))
        self.W1 = np.concatenate(
            (self.W1[:, :offset], col, self.W1[:, offset:]), axis=1
        )

        col = np.random.choice([1, -1], size=(self.k, 1))
        self.W2 = np.concatenate(
            (self.W2[:, :offset], col, self.W2[:, offset:]), axis=1
        )

    def reduce(self, Y):
        return (
            safe_sparse_dot(self.W1, Y) * safe_sparse_dot(self.W2, Y) / np.sqrt(self.k)
        )


class TensorSketchProjection(BaseProjection):

    """Tensor Sketch Projection.

    Parameters
    ----------
    k : int
        Number of reduced dimensions (i.e. rows of projection mat.).

    p : int
        Number of input dimensions (i.e. cols of projection mat.).
        ``p`` will be increased in the future due to new user/item/context insersion.
    """

    def __init__(self, k, p):
        self.k = k

        self.h1 = np.random.choice(range(k), size=p)
        self.h2 = np.random.choice(range(k), size=p)
        self.h1_indices = [np.where(self.h1 == j)[0] for j in range(k)]
        self.h2_indices = [np.where(self.h2 == j)[0] for j in range(k)]
        self.s1 = np.random.choice([1, -1], size=p)
        self.s2 = np.random.choice([1, -1], size=p)

    def insert_proj_col(self, offset):
        self.h1 = np.concatenate(
            (self.h1[:offset], np.random.choice(range(self.k), (1,)), self.h1[offset:])
        )
        self.h2 = np.concatenate(
            (self.h2[:offset], np.random.choice(range(self.k), (1,)), self.h2[offset:])
        )
        self.h1_indices = [np.where(self.h1 == j)[0] for j in range(self.k)]
        self.h2_indices = [np.where(self.h2 == j)[0] for j in range(self.k)]
        self.s1 = np.concatenate(
            (self.s1[:offset], np.random.choice([1, -1], (1,)), self.s1[offset:])
        )
        self.s2 = np.concatenate(
            (self.s2[:offset], np.random.choice([1, -1], (1,)), self.s2[offset:])
        )

    def reduce(self, Y):
        if sp.isspmatrix(Y):
            Y = Y.toarray()

        sketch1, sketch2 = self.__sketch(Y)

        return np.real(
            np.fft.ifft(
                np.fft.fft(sketch1, axis=0) * np.fft.fft(sketch2, axis=0), axis=0
            )
        )

    def __sketch(self, X):
        sketch1 = np.array(
            [
                np.sum(np.array([self.s1[idx]]).T * X[idx], axis=0)
                for idx in self.h1_indices
            ]
        )
        sketch2 = np.array(
            [
                np.sum(np.array([self.s2[idx]]).T * X[idx], axis=0)
                for idx in self.h2_indices
            ]
        )

        return sketch1, sketch2



from .base import FeatureRecommenderMixin

import time
import numpy as np

from collections import deque
from . import logger


class Evaluator(object):

    """Wrap a streaming recommender and evaluate them in the positive-only feedback setting.

    Parameters
    ----------
    recommender : RecommenderMixin
        Instance of a recommender which has already been initialized.

    repeat : bool, default=True
        Whether the same item can be repeatedly interacted by the same user.

    maxlen : int
        Size of an item buffer which stores most recently observed items.

    debug : bool, default=False
        Verbose logging if ``True``.

    References
    ----------
    .. [1] J. Vinagre et al.
           `Fast Incremental Matrix Factorization for Recommendation with Positive-only
           Feedback <http://link.springer.com/chapter/10.1007/978-3-319-08786-3_41>`_.
           In *Proc. of UMAP 2014*, pp. 459-470, July 2014.
    """

    def __init__(self, recommender, repeat=True, maxlen=None, debug=False):
        self.rec = recommender
        self.feature_rec = issubclass(recommender.__class__, FeatureRecommenderMixin)

        self.repeat = repeat

        # create a ring buffer
        # save items which are observed in most recent `maxlen` events
        self.item_buffer = deque(maxlen=maxlen)

        self.debug = debug

    def fit(self, train_events, test_events, n_epoch=1):
        """Train a model using the first 30% positive events to avoid cold-start.

        Evaluation of this batch training is done by using the next 20% positive events.
        After the batch SGD training, the models are incrementally updated by using
        the 20% test events.

        Parameters
        ----------
        train_events : list of Event
            Positive training events (0-30%).

        test_events : list of Event
            Test events (30-50%).

        n_epoch : int, default=1
            Number of epochs for the batch training.
        """
        # make initial status for batch training
        for e in train_events:
            self.__validate(e)
            self.rec.users[e.user.index]["known_items"].add(e.item.index)
            self.item_buffer.append(e.item.index)

        # for batch evaluation, temporarily save new users info
        for e in test_events:
            self.__validate(e)
            self.item_buffer.append(e.item.index)

        self.batch_update(train_events, test_events, n_epoch)

        # batch test events are considered as a new observations;
        # the model is incrementally updated based on them
        # before the incremental evaluation step
        for e in test_events:
            self.rec.users[e.user.index]["known_items"].add(e.item.index)
            self.rec.update(e)

    def evaluate(self, test_events):
        """Iterate recommend-then-update procedure and compute incremental recall.

        Parameters
        ----------
        test_events : list of Event
            Positive test events.

        Yields
        ------
        tuple
            (rank, recommend time, update time).
        """
        for i, e in enumerate(test_events):
            self.__validate(e)

            # target items (all or unobserved depending on a detaset)
            unobserved = set(self.item_buffer)
            if not self.repeat:
                unobserved -= self.rec.users[e.user.index]["known_items"]

            # item i interacted by user u must be in the recommendation candidate
            # even if it is a new item
            unobserved.add(e.item.index)

            candidates = np.asarray(list(unobserved))

            # make top-{at} recommendation for the 1001 items
            start = time.perf_counter()
            recos, scores = self.__recommend(e, candidates)
            recommend_time = time.perf_counter() - start

            rank = np.where(recos == e.item.index)[0][0]

            # Step 2: update the model with the observed event
            self.rec.users[e.user.index]["known_items"].add(e.item.index)
            start = time.perf_counter()
            self.rec.update(e)
            update_time = time.perf_counter() - start

            self.item_buffer.append(e.item.index)

            # (top-1 score, where the correct item is ranked, rec time, update time)
            yield scores[0], rank, recommend_time, update_time

    def batch_update(self, train_events, test_events, n_epoch):
        """Update a model in the batch fashion.

        Parameters
        ----------
        train_events : list of Event
            Training samples.

        test_events : list of Event
            Validation samples.

        n_epoch : int
            Number of epochs for the batch training.
        """
        for epoch in range(n_epoch):
            # SGD requires us to shuffle events in each iteration
            # * if n_epoch == 1
            #   => shuffle is not required because it is a deterministic training
            #      (i.e. matrix sketching)
            if n_epoch != 1:
                np.random.shuffle(train_events)

            # train
            for e in train_events:
                self.rec.update(e, batch_train=True)

            # test
            MPR = self.batch_evaluate(test_events)
            if self.debug:
                logger.debug("epoch %2d: MPR = %f" % (epoch + 1, MPR))

    def batch_evaluate(self, test_events):
        """Evaluate the current model in terms of Mean Percentile Rank (MPR),
        by using the given test events.

        Parameters
        ----------
        test_events : list of Event
            Validation samples to evaluate the current model.

        Returns
        -------
        float
            Mean Percentile Rank for the test set.
        """
        percentiles = np.zeros(len(test_events))

        all_items = set(self.item_buffer)
        for i, e in enumerate(test_events):

            # check if the data allows users to interact the same items repeatedly
            unobserved = all_items
            if not self.repeat:
                # make recommendation for all unobserved items
                unobserved -= self.rec.users[e.user.index]["known_items"]
                # true item itself must be in the recommendation candidates
                unobserved.add(e.item.index)

            candidates = np.asarray(list(unobserved))
            recos, scores = self.__recommend(e, candidates)

            pos = np.where(recos == e.item.index)[0][0]
            percentiles[i] = pos / (len(recos) - 1) * 100

        return np.mean(percentiles)

    def __recommend(self, e, candidates):
        if self.feature_rec:
            return self.rec.recommend(e.user, candidates, e.context)
        else:
            return self.rec.recommend(e.user, candidates)

    def __validate(self, e):
        self.__validate_user(e)
        self.__validate_item(e)

    def __validate_user(self, e):
        if self.rec.is_new_user(e.user.index):
            self.rec.register_user(e.user)

    def __validate_item(self, e):
        if self.rec.is_new_item(e.item.index):
            self.rec.register_item(e.item)



from ..base import FeatureRecommenderMixin
from ..model import OnlineSketch

import numpy as np
import numpy.linalg as ln
import scipy.sparse as sp
from sklearn import preprocessing
from sklearn.utils.extmath import safe_sparse_dot


class SketchRecommender(OnlineSketch, FeatureRecommenderMixin):

    __doc__ = OnlineSketch.__doc__

    def initialize(self):
        super(SketchRecommender, self).initialize()

    def register_user(self, user):
        super(SketchRecommender, self).register_user(user)

    def register_item(self, item):
        super(SketchRecommender, self).register_item(item)

        i_vec = item.encode(index=False, feature=True, vertical=True)
        i_vec = sp.csr_matrix(i_vec)
        if self.i_mat.size == 0:
            self.i_mat = i_vec
        else:
            self.i_mat = sp.csr_matrix(sp.hstack((self.i_mat, i_vec)))

    def update(self, e, batch_train=False):
        y = e.encode(index=False, feature=True, context=True)
        self.update_model(y)

    def score(self, user, candidates, context=np.array([])):
        # i_mat is (n_item_context, n_item) for all possible items
        # extract only target items
        i_mat = self.i_mat[:, candidates]

        n_target = len(candidates)

        # u_mat will be (n_user_context, n_item) for the target user
        u_vec = np.concatenate((user.feature, context))
        u_vec = np.array([u_vec]).T

        u_mat = sp.csr_matrix(np.repeat(u_vec, n_target, axis=1))

        # stack them into (p, n_item) matrix
        Y = sp.vstack((u_mat, i_mat))
        Y = self.proj.reduce(Y)
        Y = sp.csr_matrix(preprocessing.normalize(Y, norm="l2", axis=0))

        X = np.identity(self.k) - np.dot(self.U_r, self.U_r.T)
        A = safe_sparse_dot(X, Y, dense_output=True)

        return ln.norm(A, axis=0, ord=2)

    def recommend(self, user, candidates, context):
        scores = self.score(user, candidates, context)
        return self.scores2recos(scores, candidates)



from ..base import FeatureRecommenderMixin
from ..model import FactorizationMachine
from .. import logger

import numpy as np
import scipy.sparse as sp
from sklearn.utils.extmath import safe_sparse_dot


class FMRecommender(FactorizationMachine, FeatureRecommenderMixin):

    __doc__ = FactorizationMachine.__doc__

    def initialize(self, static=False, use_index=False):
        """Initialize a recommender by resetting stored users and items.
        Default the number of users and items to zero.

        Parameters
        ----------
        static : bool, default=False
            Disable incremental update if ``True``.

        use_index : bool, default=False
            Incorporate onehot-encoded user/item index into a feature vector.
        """
        super(FMRecommender, self).initialize()
        self.static = static
        self.use_index = use_index

    def register_user(self, user):
        super(FMRecommender, self).register_user(user)

        if self.use_index:
            n_user = self.n_user - 1

            # insert new user's row for the parameters
            self.w = np.concatenate((self.w[:n_user], np.array([0.0]), self.w[n_user:]))

            self.prev_w = np.concatenate(
                (self.prev_w[:n_user], np.array([0.0]), self.prev_w[n_user:])
            )

            rand_row = np.random.normal(0.0, 0.1, (1, self.k))

            self.V = np.concatenate((self.V[:n_user], rand_row, self.V[n_user:]))

            self.prev_V = np.concatenate(
                (self.prev_V[:n_user], rand_row, self.prev_V[n_user:])
            )

            self.p += 1

    def register_item(self, item):
        super(FMRecommender, self).register_item(item)

        n_item = self.n_item - 1

        # update the item matrix for all items
        i_vec = item.encode(
            dim=(n_item + 1), index=self.use_index, feature=True, vertical=True
        )
        sp_i_vec = sp.csr_matrix(i_vec)

        if self.i_mat.size == 0:
            self.i_mat = sp_i_vec
        elif self.use_index:
            self.i_mat = sp.vstack(
                (self.i_mat[:n_item], np.zeros((1, n_item)), self.i_mat[n_item:])
            )
            self.i_mat = sp.csr_matrix(sp.hstack((self.i_mat, sp_i_vec)))
        else:
            self.i_mat = sp.csr_matrix(sp.hstack((self.i_mat, sp_i_vec)))

        if self.use_index:

            # insert new item's row for the parameters
            h = self.n_user + n_item

            self.w = np.concatenate((self.w[:h], np.array([0.0]), self.w[h:]))

            self.prev_w = np.concatenate(
                (self.prev_w[:h], np.array([0.0]), self.prev_w[h:])
            )

            rand_row = np.random.normal(0.0, 0.1, (1, self.k))

            self.V = np.concatenate((self.V[:h], rand_row, self.V[h:]))

            self.prev_V = np.concatenate((self.prev_V[:h], rand_row, self.prev_V[h:]))
            self.p += 1

    def update(self, e, batch_train=False):
        # static baseline; w/o updating the model
        if not batch_train and self.static:
            return

        x = e.encode(index=self.use_index, n_user=self.n_user, n_item=self.n_item)

        if e.value != 1.0:
            logger.info(
                "Incremental factorization machines assumes "
                + "implicit feedback recommendation, so "
                + "the event value is automatically converted into 1.0"
            )
            e.value = 1.0

        self.update_model(x, e.value)

    def score(self, user, candidates, context=np.array([])):
        # i_mat is (n_item_context, n_item) for all possible items
        # extract only target items
        i_mat = self.i_mat[:, candidates]

        n_target = len(candidates)

        u_vec = user.encode(
            dim=self.n_user, index=self.use_index, feature=True, vertical=True
        )
        u_vec = np.concatenate((u_vec, np.array([context]).T))
        u_mat = sp.csr_matrix(np.repeat(u_vec, n_target, axis=1))

        mat = sp.vstack((u_mat, i_mat))

        # Matrix A and B should be dense (numpy array; rather than scipy CSR
        # matrix) because V is dense.
        V = sp.csr_matrix(self.V)
        A = safe_sparse_dot(V.T, mat)
        A.data[:] = A.data ** 2

        sq_mat = mat.copy()
        sq_mat.data[:] = sq_mat.data ** 2
        sq_V = V.copy()
        sq_V.data[:] = sq_V.data ** 2
        B = safe_sparse_dot(sq_V.T, sq_mat)

        interaction = (A - B).sum(axis=0)
        interaction /= 2.0  # (1, n_item); numpy matrix form

        pred = self.w0 + safe_sparse_dot(self.w, mat, dense_output=True) + interaction

        return np.abs(1.0 - np.ravel(pred))

    def recommend(self, user, candidates, context):
        scores = self.score(user, candidates, context)
        return self.scores2recos(scores, candidates)



import numpy as np


class BaseActor(object):
    def __init__(self, index, feature=np.array([])):
        self.index = index
        self.feature = feature

    def encode(self, dim=None, index=True, feature=True, vertical=False):
        """Encode an actor to an input vector for feature-based recommenders.

        Parameters
        ----------
        dim : int or None, default=None
            Number of dimensions for onehot-encoded index.
            Use ``self.index + 1`` if ``None``.

        index : bool, default=True
            Include onehot-encoded index to an input vector.

        feature : bool, default=True
            Include features associated with an actor.

        vertical : bool, default=False
            Return as a transposed n-by-1 vertical vector.

        Returns
        -------
        array
            n-dimensional vector representing an actor.
            Size can be ``(n, 1)`` or ``(1, n)``, depending on ``vertical`` parameter.

        """

        if not dim:
            dim = self.index + 1

        x = np.array([])

        if index:
            x = np.concatenate((x, self.index_one_hot(dim)))

        if feature:
            x = np.concatenate((x, self.feature))

        return x if not vertical else np.array([x]).T

    def index_one_hot(self, dim):
        """Onehot-encode own index to a vector.

        Parameters
        ----------
        dim : int
            Number of dimensions of an output vector.
            Must be greater than or equal to `self.index`.

        Returns
        -------
        array
            ``dim``-dimensional onehot-encoded vector.

        """
        if self.index >= dim:
            raise ValueError(
                "number of dimensions must be greater than index: %d" % self.index
            )

        x = np.zeros(dim)
        x[self.index] = 1.0
        return x


class User(BaseActor):

    """User object for recommenders.

    Parameters
    ----------
    index : int
        User index used as their ID. Starting from ``0``.

    features : numpy array, default=empty
        Feature vector associated with a user.
        An element can be age, gender, days from last access, etc.

    """

    def __str__(self):
        if len(self.feature) == 0:
            return "User(index={})".format(self.index)
        else:
            return "User(index={}, feature={})".format(self.index, self.feature)


class Item(BaseActor):

    """Item object for recommenders.

    Parameters
    ----------
    index : int
        Item index used as their ID. Starting from ``0``.

    features : numpy array, default=empty
        Feature vector associated with an item.
        An element can be price, category, date published, etc.

    """

    def __str__(self):
        if len(self.feature) == 0:
            return "Item(index={})".format(self.index)
        else:
            return "Item(index={}, feature={})".format(self.index, self.feature)


class Event(object):

    """An event object that represents a single user-item interaction.

    Parameters
    ----------
    user : User
        A `User` instance.

    item : Item
        An `Item` instance.

    value : float, default=1.0
        A value representing the feedback. ``1.0`` in case of positive-only feedback,
        or 5-scale value for rating prediction, for example.

    context : numpy array, default=empty
        Vector-represented contextual information associated with an interaction.
        An element can be day of the week, time, weather, etc.

    """

    def __init__(self, user, item, value=1.0, context=np.array([])):
        self.user = user
        self.item = item
        self.value = value
        self.context = context

    def encode(
        self,
        n_user=None,
        n_item=None,
        index=True,
        feature=True,
        context=True,
        vertical=False,
    ):
        """Encode an event to an input vector for feature-based recommenders.

        Parameters
        ----------
        n_user : int
            Number of users currently registered to a recommender.

        n_item : int
            Number of items currently registered to a recommender.

        index : bool, default=True
            Include onehot-encoded user/item index to an input vector.

        feature : bool, default=True
            Include features associated with user and item.

        context : bool, default=True
            Include event-specific contextual information to a vector.

        vertical : bool, default=False
            Return as a transposed n-by-1 vertical vector.

        Returns
        -------
        array
            n-dimensional vector representing a user-item interaction.
            Size can be ``(n, 1)`` or ``(1, n)``, depending on ``vertical`` parameter.

        """

        x = self.user.encode(dim=n_user, index=index, feature=feature, vertical=False)

        if context:
            x = np.concatenate((x, self.context))

        iv = self.item.encode(dim=n_item, index=index, feature=feature, vertical=False)
        x = np.concatenate((x, iv))

        assert len(x) > 0, "feature vector has zero dimension"

        return x if not vertical else np.array([x]).T

    def __str__(self):
        if len(self.context) == 0:
            return "Event(user={}, item={}, value={})".format(
                self.user, self.item, self.value
            )
        else:
            return "Event(user={}, item={}, value={}, context={})".format(
                self.user, self.item, self.value, self.context
            )



from .bprmf import BPRMFRecommender
from .factorization_machine import FMRecommender
from .matrix_factorization import MFRecommender
from .online_sketch import SketchRecommender
from .user_knn import UserKNNRecommender

__all__ = [
    "BPRMFRecommender",
    "FMRecommender",
    "MFRecommender",
    "SketchRecommender",
    "UserKNNRecommender",
]



from sklearn.base import BaseEstimator
from ..base import RecommenderMixin

import numpy as np


class Popular(BaseEstimator, RecommenderMixin):

    """Popularity-based non-personalized baseline that prioritizes
    items observed most through `update`, regardless of user profiles.
    """

    def __init__(self):
        self.freq = np.array([])

    def initialize(self):
        super(Popular, self).initialize()

    def register_user(self, user):
        super(Popular, self).register_user(user)

    def register_item(self, item):
        super(Popular, self).register_item(item)
        self.freq = np.append(self.freq, 0)

    def update(self, e, batch_train=False):
        self.freq[e.item.index] += 1

    def score(self, user, candidates):
        return self.freq[candidates]

    def recommend(self, user, candidates):
        scores = self.score(user, candidates)
        return self.scores2recos(scores, candidates, rev=True)



from .bprmf import BPRMF
from .factorization_machine import FactorizationMachine
from .matrix_factorization import MatrixFactorization
from .online_sketch import OnlineSketch
from .user_knn import UserKNN

__all__ = [
    "BPRMF",
    "FactorizationMachine",
    "MatrixFactorization",
    "OnlineSketch",
    "UserKNN",
]



from sklearn.base import BaseEstimator

import numpy as np
import scipy.sparse as sp

from .. import logger


class FactorizationMachine(BaseEstimator):

    r"""Incremental Factorization Machines (FMs).

    Parameters
    ----------
    p : int
        Number of dimensions of an input vector.

    k : int, default=40
        Number of latent factors.

    l2_reg_w0 : float, default=2.0
        :math:`\lambda_{w_0}` for L2 regularization.

    l2_reg_w : float, default=8.0
        :math:`\lambda_w` for L2 regularization.

    l2_reg_V : float, default=16.0
        :math:`\lambda_{V}` for L2 regularization.

    learn_rate : float, default=0.003
        Learning rate :math:`\eta`.

    References
    ----------
    .. [1] T. Kitazawa.
           `Incremental Factorization Machines for Persistently Cold-Starting Online
           Item Recommendation <https://arxiv.org/abs/1607.02858>`_.
           arXiv:1607.02858 [cs.LG], July 2016.
    """

    def __init__(
        self, p=None, k=40, l2_reg_w0=2.0, l2_reg_w=8.0, l2_reg_V=16.0, learn_rate=0.004
    ):

        assert p is not None

        # number of dimensions of input vectors
        self.p = p

        self.k = k
        self.l2_reg_w0 = l2_reg_w0
        self.l2_reg_w = l2_reg_w
        self.l2_reg_V = np.ones(k) * l2_reg_V
        self.learn_rate = learn_rate

        self.i_mat = sp.csr_matrix([])

        # initial parameters
        self.w0 = 0.0
        self.w = np.zeros(self.p)
        self.V = np.random.normal(0.0, 0.1, (self.p, self.k))

        # to keep the last parameters for adaptive regularization
        self.prev_w0 = self.w0
        self.prev_w = self.w.copy()
        self.prev_V = self.V.copy()

    def update_reg(self, x, err):
        x_vec = np.array([x]).T  # p x 1

        # update regularization parameters
        coeff = 4.0 * self.learn_rate * err * self.learn_rate

        self.l2_reg_w0 = max(0.0, self.l2_reg_w0 + coeff * self.prev_w0)
        self.l2_reg_w = max(0.0, self.l2_reg_w + coeff * np.inner(x, self.prev_w))

        if self.l2_reg_w0 == 0.0 or self.l2_reg_w == 0.0:
            logger.warn("reg_w0 and/or reg_w are fallen in 0.0")

        dot_v = np.dot(x_vec.T, self.V).reshape((self.k,))  # (k, )
        dot_prev_v = np.dot(x_vec.T, self.prev_V).reshape((self.k,))  # (k, )
        s_duplicated = np.dot((x_vec.T ** 2), self.V * self.prev_V).reshape(
            (self.k,)
        )  # (k, )
        self.l2_rev_V = np.maximum(
            np.zeros(self.k),
            self.l2_reg_V + coeff * (dot_v * dot_prev_v - s_duplicated),
        )

    def update_model(self, x, value):
        x_vec = np.array([x]).T  # p x 1
        interaction = (
            np.sum(np.dot(self.V.T, x_vec) ** 2 - np.dot(self.V.T ** 2, x_vec ** 2))
            / 2.0
        )
        pred = self.w0 + np.inner(self.w, x) + interaction

        # compute current error
        err = value - pred

        self.update_reg(x, err)

        # update w0 with keeping the previous value
        self.prev_w0 = self.w0
        self.w0 = self.w0 + 2.0 * self.learn_rate * (
            err * 1.0 - self.l2_reg_w0 * self.w0
        )

        # keep the previous w for auto-parameter optimization
        self.prev_w = np.empty_like(self.w)
        self.prev_w[:] = self.w

        # keep the previous V
        self.prev_V = np.empty_like(self.V)
        self.prev_V[:] = self.V

        # update w and V
        prod = np.dot(np.array([x]), self.prev_V)  # (1, p) and (p, k) => (1, k)
        for pi in range(self.p):
            if x[pi] == 0.0:
                continue

            self.w[pi] = self.prev_w[pi] + 2.0 * self.learn_rate * (
                err * x[pi] - self.l2_reg_w * self.prev_w[pi]
            )

            g = err * x[pi] * (prod - x[pi] * self.prev_V[pi])
            self.V[pi] = self.prev_V[pi] + 2.0 * self.learn_rate * (
                g - self.l2_reg_V * self.prev_V[pi]
            )



import flurs

VERSION = flurs.__version__

DISTNAME = "flurs"
DESCRIPTION = "A library for streaming recommendation algorithms"
with open("README.rst") as f:
    LONG_DESCRIPTION = f.read()
LONG_DESCRIPTION_CONTENT_TYPE = "text/x-rst"
AUTHOR = "Takuya Kitazawa"
AUTHOR_EMAIL = "k.takuti@gmail.com"
MAINTAINER = AUTHOR
MAINTAINER_EMAIL = AUTHOR_EMAIL
LICENSE = "MIT"
URL = "https://github.com/takuti/flurs"
DOWNLOAD_URL = "https://pypi.org/project/flurs/#files"


def setup_package():
    from setuptools import setup, find_packages

    metadata = dict(
        name=DISTNAME,
        version=VERSION,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        license=LICENSE,
        url=URL,
        download_url=DOWNLOAD_URL,
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "License :: OSI Approved",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering",
            "Development Status :: 4 - Beta",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: Unix",
            "Operating System :: MacOS",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
        packages=find_packages(exclude=["*tests*"]),
        install_requires=[
            "numpy>=1.14.6",
            "scipy>=1.6.3",
            "scikit_learn>=1.0",
            "mmh3>=3.0.0",
        ],
        extras_require={
            "docs": [
                "sphinx_rtd_theme",
                "sphinx-gallery",
            ],
            "tests": {
                "pytest>=5.0.1",
                # https://github.com/pytest-dev/pytest/issues/4608
                "pytest-remotedata==0.3.2",
                "flake8>=3.8.2",
                "black>=21.6b0",
                "mypy>=0.770",
                "pre-commit",
            },
        },
    )

    setup(**metadata)


if __name__ == "__main__":
    setup_package()



from .data.entity import User, Item

import numpy as np


class RecommenderMixin(object):

    """Mixin injected into a model to make it a recommender."""

    def initialize(self, *args):
        """Initialize a recommender by resetting stored users and items.
        Default the number of users and items to zero.
        """
        # number of observed users
        self.n_user = 0

        # store user data
        self.users = {}

        # number of observed items
        self.n_item = 0

        # store item data
        self.items = {}

    def is_new_user(self, u):
        """Check if a user is already registered to a recommender.

        Parameters
        ----------
        u : int
            User index.

        Returns
        -------
        bool
            Whether the user is new.
        """
        return u not in self.users

    def register(self, actor):
        """Register a user or item to a recommender.
        Delegate the process to `register_user` or `register_item`.

        Parameters
        ----------
        actor : User or Item
            A `User` or `Item` instance to register.
        """
        t = type(actor)
        if t == User:
            self.register_user(actor)
        elif t == Item:
            self.register_item(actor)

    def register_user(self, user):
        """Register a user to a recommender with an empty dictionary that
        records a set of item indices observed in the past.

        Parameters
        ----------
        user : User
            A `User` instance to register.
        """
        self.users[user.index] = {"known_items": set()}
        self.n_user += 1

    def is_new_item(self, i):
        """Check if an item is already registered to a recommender.

        Parameters
        ----------
        i : int
            Item index.

        Returns
        -------
        bool
            Whether the item is new.
        """
        return i not in self.items

    def register_item(self, item):
        """Register an item to a recommender with an empty dictionary that is
        potentially used to record any auxiliary information.

        Parameters
        ----------
        item : Item
            A `Item` instance to register.
        """
        self.items[item.index] = {}
        self.n_item += 1

    def update(self, e, batch_train):
        """Update model parameters given a single user-item interaction.

        Parameters
        ----------
        e : Event
            Observed event representing a user-item interaction.

        batch_train : bool
            Let recommender know if the update operation is part of batch training
            rather than incremental, online update.
        """
        pass

    def score(self, user, candidates):
        """Compute scores for the pairs of given user and item candidates.

        Parameters
        ----------
        user : User
            Target user.

        candidates : numpy array, (# candidates, )
            Target item indices.

        Returns
        -------
        array, (# candidates, )
            Predicted values for the given user-candidates pairs.
        """
        return

    def recommend(self, user, candidates):
        """Recommend items for a user by calculating sorted list of item candidates.

        1. Scores are computed.
        2. `scores2recos()` is called to convert the scores into a recommendation list.

        Parameters
        ----------
        user : User
            Target user.

        candidates : numpy array, (# target items, )
            Target items indices.
            Only these items are considered as the recommendation candidates.

        Returns
        -------
        (array, array)
            A tuple of ``(sorted list of item indices, sorted scores)``.
        """
        return

    def scores2recos(self, scores, candidates, rev=False):
        """Get recommendation list for a user based on scores.

        Parameters
        ----------
        scores : numpy array, (n_target_items,)
            Scores for the target items. Smaller score indicates a promising item.

        candidates : numpy array, (# target items, )
            Target items indices.
            Only these items are considered as the recommendation candidates.

        rev : bool, default=False
            If ``True``, sort and return items in an descending order.
            The default is an ascending order (i.e., smaller scores are more promising).

        Returns
        -------
        (array, array)
            A tuple of ``(sorted list of item indices, sorted scores)``.
        """
        sorted_indices = np.argsort(scores)

        if rev:
            sorted_indices = sorted_indices[::-1]

        return candidates[sorted_indices], scores[sorted_indices]


class FeatureRecommenderMixin(RecommenderMixin):

    """Mixin injected into a model to make it a feature-based recommender."""

    def score(self, user, candidates, context):
        """Compute scores for the pairs of given user and item candidates.

        Parameters
        ----------
        user : User
            Target user.

        candidates : numpy array, (# candidates, )
            Target item indices.

        context : (numpy array)
            Feature vector representing contextual information.

        Returns
        -------
        array, (# candidates, )
            Predicted values for the given user-candidates pairs.
        """
        return

    def recommend(self, user, candidates, context):
        """Recommend items for a user by calculating sorted list of item candidates.

        1. Scores are computed.
        2. `scores2recos()` is called to convert the scores into a recommendation list.

        Parameters
        ----------
        user : User
            Target user.

        candidates : numpy array, (# target items, )
            Target items indices.
            Only these items are considered as the recommendation candidates.

        context : (numpy array)
            Feature vector representing contextual information.

        Returns
        -------
        (array, array)
            A tuple of ``(sorted list of item indices, sorted scores)``.
        """
        return



from sklearn.base import BaseEstimator

import numpy as np


class UserKNN(BaseEstimator):

    """Incremental User-based Collaborative Filtering using k-Nearest-Neighbor (kNN).

    Parameters
    ----------
    k : int, default=5
        Number of nearest neighbors.

    References
    ----------
    .. [1] M. Pepagelis et al. **Incremental Collaborative Filtering for
           Highly-Scalable Recommendation Algorithms**. In *Foundations of Intelligent
           Systems*, pp. 553-561, Springer Berlin Heidelberg, 2005.
    """

    def __init__(self, k=5):
        # number of nearest neighbors
        self.k = k

        # user-item matrix
        self.R = np.array([])

        # user-user similarity matrix
        self.S = np.array([])

        # user-user similarity: S = B / (sqrt(C) * sqrt(D))
        self.B = np.array([])
        self.C = np.array([])
        self.D = np.array([])

    def update_model(self, ua, ia, value):
        prev_r = self.R[ua, ia]
        new_submit = prev_r == 0
        self.R[ua, ia] = value

        prev_mean = self.users[ua]["mean"]

        if new_submit:
            self.users[ua]["count"] += 1
            self.users[ua]["mean"] = (
                self.R[ua, ia] / self.users[ua]["count"]
                + (self.users[ua]["count"] - 1) / self.users[ua]["count"] * prev_mean
            )
        else:
            self.users[ua]["mean"] = (self.R[ua, ia] - prev_r) / (
                self.users[ua]["count"] - 1
            ) + prev_mean

        d = self.users[ua]["mean"] - prev_mean

        for uy in range(self.n_user):
            # skip myself
            if uy == ua:
                continue

            e = f = g = 0.0

            had_uy_rated_ia = self.R[uy, ia] != 0
            if had_uy_rated_ia:
                ua_normalized = self.R[ua, ia] - self.users[ua]["mean"]
                uy_normalized = self.R[uy, ia] - self.users[uy]["mean"]

                if new_submit:
                    e = ua_normalized * uy_normalized
                    f = ua_normalized ** 2
                    g = uy_normalized ** 2
                else:
                    e = (self.R[ua, ia] - prev_r) * uy_normalized
                    f = (self.R[ua, ia] - prev_r) ** 2 + 2 * (
                        self.R[ua, ia] - prev_r
                    ) * ua_normalized
                    g = 0.0

            for ih in range(self.n_item):
                # only for co-rated items
                if self.R[ua, ih] != 0 and self.R[uy, ih] != 0:
                    e = e - d * (self.R[uy, ih] - self.users[uy]["mean"])
                    f = f + d ** 2 - 2 * d * (self.R[ua, ih] - prev_mean)

            self.B[ua, uy] += e
            self.C[ua, uy] += f
            self.D[ua, uy] += g

        # avoid zero division
        idx = (self.C[ua, :] != 0) & (self.C[ua, :] != 0)
        self.S[ua, idx] = self.B[ua, idx] / (
            np.sqrt(self.C[ua, idx]) * np.sqrt(self.D[ua, idx])
        )
